//
// Iff2Bmp v0.3 (D) by ssjx (http://ssjx.co.uk)
//
// Compile with:
//		dmd iff2bmp.c
//
//
// Version 0.3 (21/05/2023)
// + (07/05/2023) File not found message!
// + (21/05/2023) Uses seek to skip unknown chunks. Faster, one less allocation.
//
// Version 0.2 (01/01/2023)
// + HAM6 / EHB added
// + Fixed problem in width function (used % instead of &)
// + Fixed fault in HAM6 (did not follow spec correctly)
//
// Version 0.1a (28/12/2022)
// + No changes, fresh recompile with dmd v2.101.1
//
// Version 0.1 (24/10/2020)
//

import std.stdio;
import std.File;

import std.bitmanip;	// swapEndian

// Bitmap Header - 14 bytes (NOT 16)
struct my_bmp_header{
	align(1):
	char[2] type;
	uint size;
	ushort res1;
	ushort res2;
	uint offset;
};

// 40 bytes
struct my_bmp_info{
	align(1):
	uint size;
	uint width;
	uint height;
	ushort planes;
	ushort depth;
	uint comp;
	uint isize;
	uint xppm;
	uint yppm;
	uint clrused;
	uint clrimp;
};

//
// IFF Structs
//

struct my_iff_header{
	align(1):
	char[4] chunk_id; 	// 	FORM
	uint size;
	char[4] format_id; 	//	ILBM
};

struct my_chunk_header{
	align(1):
	char[4] chunk_id; 	//BMHD
	uint size;
};

struct my_bmhd_header{
	align(1):
	ushort w;
	ushort h;
	ushort xorigin;
	ushort yorigin;
	ubyte planes;
	ubyte mask;
	ubyte comp;
	char pad1;
	ushort transclr;
	ubyte xaspect;
	ubyte yaspect;
	ushort pagew;
	ushort pageh;
};

struct rgb{
	ubyte r;
	ubyte g;
	ubyte b;
};


rgb[256] MY_RGB;	//0 - 255
bool HAM=false;	
bool EHB=false;	

//
//
//

@safe int needed(int w){
	int c=0; 
	
	while(w>0){
		w-=8;
		c=c+1;
	}
	
	// So row has even number of bytes
	if ((c & 1)==1){c++;}	// %
	
	//writeln("IFF Bytes needed per row: %d",c);
	return c;
}

// Actual BMP row with (including padding byte) 24bit
@safe int byte_width(const int w){
	int bmp_row_width=w*3;
	
	int b=0;
	while(1){
		b+=4;
		if (b>=bmp_row_width){break;}
	}
	
	//writeln("BMP Bytes needed per row: %d",b);
	return b;
}

//
@safe void cydraw(const ubyte[] from,ubyte[] to,int width,uint height, uint planes){
	uint ina=0;
	uint inb;
	uint inc;
	
	ubyte a,b,c,d,e,f,g,h;
	
	ubyte ar,ag,ab;
	
	uint x,xa;
	
	ubyte prev_r;
	ubyte prev_g;
	ubyte prev_b;
	
	ubyte[8] pix;	// Planes
	ubyte[8] green;	// Planes
	ubyte[8] blue;	// Planes
	
	immutable uint iff_row_width=needed(width);
	immutable uint bmp_row_width=byte_width(width);
	
	uint pos;
	
	foreach(int j;0..height){
		inb=ina;
		x=0;
		
		prev_r=0;
		prev_g=0;
		prev_b=0;
		
		foreach(int i;0..iff_row_width){
			
			// Clear plane arrays
			pix[]=0;
			green[]=0;
			blue[]=0;
						
			inc=inb;
			if (planes<24){
				foreach(int depth;0..planes){
					h=from[inc] & 0x01;
					g=(from[inc] & 0x02)>>1;
					f=(from[inc] & 0x04)>>2;
					e=(from[inc] & 0x08)>>3;
					d=(from[inc] & 0x10)>>4;
					c=(from[inc] & 0x20)>>5;
					b=(from[inc] & 0x40)>>6;
					a=(from[inc] & 0x80)>>7;
					
					pix[0]+=a<<depth;
					pix[1]+=b<<depth;
					pix[2]+=c<<depth;
					pix[3]+=d<<depth;
					pix[4]+=e<<depth;
					pix[5]+=f<<depth;
					pix[6]+=g<<depth;
					pix[7]+=h<<depth;
					inc+=iff_row_width;
				}
			}else{
				foreach(int depth;0..8){
					h=from[inc] & 0x01;
					g=(from[inc] & 0x02)>>1;
					f=(from[inc] & 0x04)>>2;
					e=(from[inc] & 0x08)>>3;
					d=(from[inc] & 0x10)>>4;
					c=(from[inc] & 0x20)>>5;
					b=(from[inc] & 0x40)>>6;
					a=(from[inc] & 0x80)>>7;
					// Reusing pix array to hold red
					pix[0]+=a<<depth;
					pix[1]+=b<<depth;
					pix[2]+=c<<depth;
					pix[3]+=d<<depth;
					pix[4]+=e<<depth;
					pix[5]+=f<<depth;
					pix[6]+=g<<depth;
					pix[7]+=h<<depth;
					inc+=iff_row_width;
				}
				
				foreach(int depth;8..16){
					h=from[inc] & 0x01;
					g=(from[inc] & 0x02)>>1;
					f=(from[inc] & 0x04)>>2;
					e=(from[inc] & 0x08)>>3;
					d=(from[inc] & 0x10)>>4;
					c=(from[inc] & 0x20)>>5;
					b=(from[inc] & 0x40)>>6;
					a=(from[inc] & 0x80)>>7;
					
					green[0]+=a<<(depth-8);
					green[1]+=b<<(depth-8);
					green[2]+=c<<(depth-8);
					green[3]+=d<<(depth-8);
					green[4]+=e<<(depth-8);
					green[5]+=f<<(depth-8);
					green[6]+=g<<(depth-8);
					green[7]+=h<<(depth-8);
					inc+=iff_row_width;
				}
				
				foreach(int depth;16..24){
					h=from[inc] & 0x01;
					g=(from[inc] & 0x02)>>1;
					f=(from[inc] & 0x04)>>2;
					e=(from[inc] & 0x08)>>3;
					d=(from[inc] & 0x10)>>4;
					c=(from[inc] & 0x20)>>5;
					b=(from[inc] & 0x40)>>6;
					a=(from[inc] & 0x80)>>7;
					
					blue[0]+=a<<(depth-16);
					blue[1]+=b<<(depth-16);
					blue[2]+=c<<(depth-16);
					blue[3]+=d<<(depth-16);
					blue[4]+=e<<(depth-16);
					blue[5]+=f<<(depth-16);
					blue[6]+=g<<(depth-16);
					blue[7]+=h<<(depth-16);
					inc+=iff_row_width;
				}
			}
			
			foreach(int row;0..8){
				
				if (planes<24){
					ar=MY_RGB[pix[row]].r;
					ag=MY_RGB[pix[row]].g;
					ab=MY_RGB[pix[row]].b;
					
					if (HAM==true){
				
						final switch(planes){
							case 6:
								if ((pix[row] & 0x30)==0x20){ 
									// red
									ar=(pix[row] & 0x0f)<<4;
									ar=ar | (pix[row] & 0x0f);
									ag=prev_g;
									ab=prev_b;
								}
								
								if ((pix[row] & 0x30)==0x30){
									// blue
									ar=prev_r;
									ag=(pix[row] & 0x0f)<<4;
									ag=ag | (pix[row] & 0x0f);
									ab=prev_b;
								}
							
								if ((pix[row] & 0x30)==0x10){ 
									// green
									ar=prev_r;
									ag=prev_g;
									ab=(pix[row] & 0x0f)<<4;
									ab=ab | (pix[row] & 0x0f);
								}
							break;
							
							case 8:
								if ((pix[row] & 0xc0)==128){ 
									// red
									ar=(pix[row] & 0x3f)<<2;
									ag=prev_g;
									ab=prev_b;
								}
								
								if ((pix[row] & 0xc0)==192){
									// blue
									ar=prev_r;
									ag=(pix[row] & 0x3f)<<2;
									ab=prev_b;
								}
							
								if ((pix[row] & 0xc0)==64){ 
									// green
									ar=prev_r;
									ag=prev_g;
									ab=(pix[row] & 0x3f)<<2;
								
								}
							break;
						}
					
						prev_r=ar;
						prev_g=ag;
						prev_b=ab;
					}
					
				}else{
					ar=pix[row];
					ag=green[row];
					ab=blue[row];
				}
				
				// Write pixel
				pos=(((height-j)-1)*bmp_row_width)+(x*3);
				
				to[pos]=ab;
				to[pos+1]=ag;
				to[pos+2]=ar;
				//
				
				x=x+1;
				if (x==width){break;}
			}
				
			inb+=1;
			if (x==width){break;}
		}
		
		ina+=(iff_row_width*planes);
	}
}


//

void main(string[] args){
	
	writeln("Iff2Bmp v0.3 D (21/05/23) by ssjx (http://ssjx.co.uk)");
	writeln("------------------------------------------------------");
	
	if (args.length!=2){
		writeln("Can convert an Amiga 8bit, 24bit, EHB or HAM6/8 IFF ILBM image to a Windows BMP file.");
		writeln("Usage: iff2bmp myfile.iff");
		return;
	}
	string fn=args[1];
	
	if (exists(fn)==false){
		writeln("File not found!");
		writeln("Usage: iff2bmp myfile.iff");
		return;
	}
	
	writeln("Loading...");
		
	my_iff_header[1] iff_header;

	// Main IFF header
	File file = File(fn, "rb");
	file.rawRead(iff_header);
	
	if (iff_header[0].format_id!="ILBM"){
		writeln("This does not look like an Amiga IFF ILBM file...");
		return;
	}

	//
	// Got this far then probably a valid file...
	//
	
	my_chunk_header[1] chunk_header;
	my_bmhd_header[1] bmhd_header;
	
	uint framesize=0;
	uint bmpsize;
	bool done=false;
	
	ubyte[] frame;
	
	while(1){
		file.rawRead(chunk_header);
	
		chunk_header[0].size=chunk_header[0].size.swapEndian;
		//writeln("| Size  : %d",chunk_header.size);
		
		switch(chunk_header[0].chunk_id){
			case "BMHD":
				file.rawRead(bmhd_header);
				
				bmhd_header[0].w=bmhd_header[0].w.swapEndian;
				bmhd_header[0].h=bmhd_header[0].h.swapEndian;
					
				bmhd_header[0].pagew=bmhd_header[0].pagew.swapEndian;
				bmhd_header[0].pageh=bmhd_header[0].pageh.swapEndian;
				
				immutable int row_width=needed(bmhd_header[0].w);
				bmpsize=row_width*(bmhd_header[0].h)*(bmhd_header[0].planes);
				
				writeln("Loading Bitmap Header");
				//writeln("--> Size  : %d",chunk_header.size);
				writeln("--> Dimensions  : ",bmhd_header[0].w," x ", bmhd_header[0].h," x ",bmhd_header[0].planes);
				writeln("--> Page  : ",bmhd_header[0].pagew," x ",bmhd_header[0].pageh);
				writeln("--> Compression  : ",bmhd_header[0].comp);
				writeln("--> Mask  : ",bmhd_header[0].mask);
				writeln("--> Pic Bytes  : ",bmpsize);
			break;
			
			case "CAMG":
				writeln("Loading CAMG");
				//writeln("--> Size  : %d",my_chunk_header.size);
				
				uint[1] viewport;
				
				file.rawRead(viewport);
				viewport[0]=viewport[0].swapEndian;
				
				//print"--> Viewport "+bin(viewport)
				if ((viewport[0] & 0x80)==0x80){
					writeln("--> Extra Half Bright");
					EHB=true;
					}
				if ((viewport[0] & 0x800)==0x800){
					writeln("--> Hold And Modify (HAM)");
					HAM=true;
				}
			break;
			
			case "CMAP":
				writeln("Loading Colour Map");
				//writeln("--> Size  : %d",my_chunk_header.size);
				
				immutable int numcols=chunk_header[0].size/3;
				writeln( "--> Colours  : ",numcols);
							
				ubyte[] my_rgb_array;
				my_rgb_array.length=chunk_header[0].size;
				
				file.rawRead(my_rgb_array);
			
				int a=0;
				foreach(int c;0..numcols){
					MY_RGB[c].r=my_rgb_array[a];
					MY_RGB[c].g=my_rgb_array[a+1];
					MY_RGB[c].b=my_rgb_array[a+2];
					a+=3;
				}
				
				if (EHB==true){
					a=0;
					foreach(int c;0..numcols){
						MY_RGB[c+32].r=my_rgb_array[a]>>1;
						MY_RGB[c+32].g=my_rgb_array[a+1]>>1;
						MY_RGB[c+32].b=my_rgb_array[a+2]>>1;
						a+=3;
					}
				}
				
			break;
			
			case "BODY":
				writeln("Loading BODY");
				//writeln("--> Size  : %d",my_chunk_header.size);
				
				framesize=chunk_header[0].size;
				frame.length=framesize;
				file.rawRead(frame);
				
				done=true;
			break;
			
			default:
				writeln("Unknown Chunk! ",chunk_header[0].chunk_id);	
				file.seek(chunk_header[0].size,SEEK_CUR);
			break;
		}
		
		if (done==true){break;}
	}

	if (bmhd_header[0].comp>1){
		writeln("Unknown compression method so aborting!");
		return;
	}
	
	writeln("Converting...");
	
	// Decompress IFF to picture buffer
	ubyte[] iff_data;
	iff_data.length=bmpsize;
	
	switch (bmhd_header[0].comp){
		case 0:
			writeln("Not compressed!");
			iff_data[]=frame[];
		break;
				
		case 1:
			writeln("Decompressing!");
			uint ps=0;
			uint op=0;
			int exit=0;
			
			ubyte value;
			ubyte outcol;
			
			while(true){
				value=frame[ps];
				ps=ps+1;
				if (ps<framesize){
				
					if (value<=127){
						for(int i=0;i<(value+1);i++){
							iff_data[op]=frame[ps];
							ps=ps+1;
							op=op+1;
						}
					}
						
					if (value==128){
						writeln("End byte found! ",ps," / ",framesize);
						exit=1;
					}
						
					if (value>128){
						outcol=frame[ps];
						ps=ps+1;
					
						for(int i=0;i<((255-value)+2);i++){ 
							iff_data[op]=outcol;
							op=op+1;
						}
					}	
					
				}else{
					writeln("Exit due to size ",ps," / ",framesize);
					exit=2;
				}
			
				if (ps>=framesize){break;}
				if (exit>0){break;}
			}
			
			writeln("Uncompressed bytes ",op," should be near ",bmpsize);
		break;
		
		default:
			// Should never reach this...
			writeln("Unknown compression method so aborting!");
			return;
		break;
	}
		
	// Make a BMP Header (Just the essentials so it opens!)
	my_bmp_header[1] bmp_header; 
	
	bmp_header[0].type[0]='B';
	bmp_header[0].type[1]='M';
	
	bmp_header[0].offset=54;
	
	my_bmp_info[1] bmp_info; 
	
	bmp_info[0].size=40;
	bmp_info[0].width=bmhd_header[0].w;
	bmp_info[0].height=bmhd_header[0].h;
	
	bmp_info[0].planes=1;
	bmp_info[0].depth=24;
	
	// Find out byte width (including padding) needed
	immutable uint bmp_row_width=byte_width(bmp_info[0].width);
	immutable uint bmp_size=(bmp_row_width*bmp_info[0].height);
		
	ubyte[] bmp_data;
	bmp_data.length=bmp_size;
		
	// Convert Amiga planar data to Windows chunky
	cydraw(iff_data, bmp_data,bmp_info[0].width,bmp_info[0].height,bmhd_header[0].planes);
	
	string outfn=fn~".bmp";	// <-- Note squiggle used to join strings!

	File outfile = File(outfn, "wb");   	
	writeln("Saving as ",outfn,"...");
	
	outfile.rawWrite(bmp_header);
	outfile.rawWrite(bmp_info);
	outfile.rawWrite(bmp_data);
	outfile.close();		
	
}


